import logging
from typing import List

from edpconfigprovider import ConfigProvider
from edpconfigprovider.exceptions import NoConfigurationException

logger = logging.getLogger(__name__)


class FallbackConfigProvider(ConfigProvider):
    def __init__(self, providers: List[ConfigProvider]) -> None:
        if not providers:
            raise ValueError("FallbackConfigProvider requires at least one provider")
        self._providers = providers

    def get_configuration(self) -> dict:
        last_exception = NoConfigurationException("fallback")
        for i, provider in enumerate(self._providers):
            try:
                return provider.get_configuration()
            except NoConfigurationException as e:
                logger.warning(
                    "Provider %s (%d/%d) failed: %s. Trying next.",
                    type(provider).__name__, i + 1, len(self._providers), e
                )
                last_exception = e

        raise last_exception



import json
from pathlib import Path
from typing import (
    Dict,
    Union,
)

from edpconfigprovider import ConfigProvider
from edpconfigprovider.exceptions import NoConfigurationException


class JsonConfigProvider(ConfigProvider):
    def __init__(self, json_path: Union[str, Path, Dict]) -> None:
        self.json_path = json_path

    def get_configuration(self) -> dict:
        if isinstance(self.json_path, dict):
            return self.json_path

        if isinstance(self.json_path, str):
            try:
                return json.loads(self.json_path)
            except json.JSONDecodeError:
                pass

        configs_files = [self.json_path]

        if isinstance(self.json_path, str):
            configs_files = self.json_path.split(",")

        json_config = {}
        for config_file in configs_files:
            try:
                file_config = self._read_json_from_file_as_dictionary(config_file)
                json_config.update(file_config)
            except FileNotFoundError:
                raise NoConfigurationException(configuration_name=str(config_file))

        return json_config

    @staticmethod
    def _read_json_from_file_as_dictionary(path: Union[str, Path]) -> dict:
        with open(path, "rb") as f:
            return json.loads(f.read().decode("utf-8"))



import os
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Optional,
    TypeVar,
    Union,
)

from edpconfigprovider.builder import (
    ProviderBuilder,
    ProviderBuilderFactory,
)
from edpconfigprovider.util import (
    ByTenantDynamicConfigProvider,
    MultiConfigProvider,
    ConfigMapper,
)

T = TypeVar("T")



def build_multi_tenant_app_config_provider(
    env_name: str,
    sub_env_name: str,
    application_name: Optional[str],
    transformer_method: Callable[[Dict], T],
) -> ConfigMapper[T]:

    full_app_name = f"backoffice-{env_name}-{sub_env_name}"
    env_builder = ProviderBuilderFactory.get_app_config_provider_builder(
        application_name=full_app_name,
        profile_name="env",
        environment_name=env_name,
    )
    tenant_builder = ProviderBuilderFactory.get_by_tenant_app_config_provider_builder(
        application_name=full_app_name,
        profile_name_template="tenant-{tenant_name}",
        environment_name=env_name,
    )

    dynamic_builder = ByTenantDynamicConfigProvider(provider_builder=tenant_builder)
    app_builder: Optional[ProviderBuilder] = None
    if application_name is not None:
        app_builder = ProviderBuilderFactory.get_app_config_provider_builder(
            application_name=full_app_name,
            profile_name=f"app-{application_name}",
            environment_name=env_name,
        )

    providers = {
        "env": env_builder.build_provider(),
        "tenant": dynamic_builder,
    }
    if app_builder is not None:
        providers["app"] = app_builder.build_provider()

    multi = MultiConfigProvider(providers=providers)

    return ConfigMapper(
        config_provider=multi,
        transform_method=transformer_method,
    )

def build_json_config_provider(
    env_config_path: Union[str, Path],
    tenant_config_path: Optional[Union[str, Path]],
    app_config_path: Optional[Union[str, Path]],
    transformer_method: Callable[[Dict], T],
) -> ConfigMapper:
    env_config_provider = ProviderBuilderFactory.get_json_provider_builder(json_path=env_config_path).build_provider()
    dynamic_builder = None
    if tenant_config_path is not None:
        tenant_builder = ProviderBuilderFactory.get_by_tenant_json_provider_builder(json_path=tenant_config_path)
        dynamic_builder = ByTenantDynamicConfigProvider(provider_builder=tenant_builder)
    app_config_provider = None
    if app_config_path is not None:
        app_config_provider = ProviderBuilderFactory.get_json_provider_builder(json_path=app_config_path).build_provider()

    providers = {
        "env": env_config_provider,
    }

    if dynamic_builder is not None:
        providers["tenant"] = dynamic_builder

    if app_config_provider is not None:
        providers["app"] = app_config_provider

    multi = MultiConfigProvider(providers=providers)

    return ConfigMapper(
        config_provider=multi,
        transform_method=transformer_method,
    )


def build_multi_tenant_app_config_provider_from_env(
    transformer_method: Callable[[Dict], T],
):
    return build_multi_tenant_app_config_provider(
        env_name=os.environ["APP_CONFIG_ENV_NAME"],
        sub_env_name=os.environ["APP_CONFIG_CLUSTER_NAME"],
        application_name=os.environ.get("APP_CONFIG_APPLICATION_NAME"),
        transformer_method=transformer_method,
    )

def build_json_config_provider_from_env(
    transformer_method: Callable[[Dict], T],
):
    return build_json_config_provider(
        env_config_path=os.environ["ENV_CONFIG_PATH"],
        tenant_config_path=os.environ.get("TENANT_CONFIG_PATH"),
        app_config_path=os.environ.get("APP_CONFIG_PATH"),
        transformer_method=transformer_method,
    )

def build_config_provider_from_env(
    transformer_method: Callable[[Dict], T],
):
    if os.environ.get("ENV_CONFIG_PATH") is not None:
        return build_json_config_provider_from_env(transformer_method=transformer_method)
    elif os.environ.get("APP_CONFIG_ENV_NAME") is not None and os.environ.get("APP_CONFIG_CLUSTER_NAME") is not None:
        return build_multi_tenant_app_config_provider_from_env(transformer_method=transformer_method)
    else:
        raise ValueError("No valid configuration environment variables found")



import json
import logging
from datetime import (
    UTC,
    datetime,
    timedelta,
)

import boto3
from botocore.exceptions import ClientError

from edpconfigprovider import ConfigProvider
from edpconfigprovider.exceptions import NoConfigurationException

logger = logging.getLogger(__name__)


class AwsAppConfigProvider(ConfigProvider):
    def __init__(
        self,
        application_name: str,
        profile_name: str,
        environment_name: str,
    ):
        self.application_name = application_name
        self.profile_name = profile_name
        self.environment_name = environment_name

        self._client = boto3.client("appconfigdata")
        self._configuration_token = None
        self._next_poll_time = None
        self._last_config = None

    def get_configuration(self) -> dict:
        if self._next_poll_time is not None and datetime.now(tz=UTC) < self._next_poll_time:
            return self._last_config

        try:
            self._ensure_session()
            response = self._client.get_latest_configuration(ConfigurationToken=self._configuration_token)
            self._configuration_token = response["NextPollConfigurationToken"]
            self._next_poll_time = datetime.now(tz=UTC) + timedelta(seconds=response["NextPollIntervalInSeconds"])

            new_config = response["Configuration"].read()
            if new_config != b"":
                self._last_config = json.loads(new_config)
            elif self._last_config is None:
                config_name = self._get_config_name()
                logger.error("No configuration found for %s", config_name)
                raise NoConfigurationException(configuration_name=config_name)

            return self._last_config
        except ClientError:
            config_name = self._get_config_name()
            logger.exception("Error getting configuration from AWS AppConfig: %s", config_name)

            self._configuration_token = None
            raise NoConfigurationException(configuration_name=config_name)

    def _ensure_session(self):
        if self._configuration_token is None:
            session = self._client.start_configuration_session(
                ApplicationIdentifier=self.application_name,
                ConfigurationProfileIdentifier=self.profile_name,
                EnvironmentIdentifier=self.environment_name,
                RequiredMinimumPollIntervalInSeconds=15,
            )
            self._configuration_token = session["InitialConfigurationToken"]

    def _get_config_name(self):
        return f"{self.application_name}/{self.environment_name}/{self.profile_name}"



from ._ByTenantDynamicConfigProvider import ByTenantDynamicConfigProvider
from ._FallbackConfigProvider import FallbackConfigProvider
from ._MultiConfigProvider import MultiConfigProvider
from ._ConfigMapper import ConfigMapper
